#include "lcompStreamDevice.h"
#include "lcomp_headers.h"
#include "../../lqmeas_errs.h"

double lcompStreamDevice::adcFreq() const {
    return m_adcCfg.FreqAdc;
}

double lcompStreamDevice::adcChFreq() const {
    return m_adcCfg.FreqCh <= 0 ? m_adcCfg.FreqAdc/adcEnabledChCnt() : m_adcCfg.FreqCh ;
}

bool lcompStreamDevice::adcChEnabled(int ch) const {
    return m_adcCfg.Ch[ch].Enabled;
}

double lcompStreamDevice::adcChRange(int ch) const {
    return adcRangeMaxVal(m_adcCfg.Ch[ch].Range);
}

LQMeasError lcompStreamDevice::adcSetConfig(lcompStreamDevice::AdcConfig cfg) {
    return adcSetRawConfig(&cfg, sizeof(cfg));
}

LQMeasError lcompStreamDevice::adcGetData(double *data, unsigned size, unsigned flags, unsigned tout, unsigned *recvd_size) {
    return m_adcReceiver.getData(this, data, size, flags, tout, recvd_size);
}

LQMeasError lcompStreamDevice::lcompProcessRecvdData(const void *src_buf, unsigned size, double *dst_buf, unsigned flags, unsigned offs) {
    const qint16 *rcv_wrds = (const qint16 *)src_buf;  
    unsigned ch_cnt = adcEnabledChCnt();
    for (unsigned i=0; i < size; i++) {
        dst_buf[i] = ((double)rcv_wrds[i])*adcChRange(adcChNum((offs + i) % ch_cnt))/8000;
    }
    return LQMeasError::Success();
}

lcompStreamDevice::lcompStreamDevice(IDaqLDevice *dev, const __SLOT_PARAM *par, unsigned slot,
                                     LQMeasAdc::AdcType adcType) : lcompDevice(dev, par, slot), LQMeasAdc(adcType) {
    adcSetChCnt(LCOMP_SEQ_COM_CH_CNT);
    m_adcReceiver.moveToThread(&m_adcReceiver);
}

LQMeasError lcompStreamDevice::protAdcStart() {
    LQMeasError err = LQMeasError::Success();
    ADC_PAR adcPar;

    err = fillAdcParams(&adcPar);

    if (err.isSuccess()) {
        m_dev->EnableCorrection(1);
        err = m_adcReceiver.setStreamParams(m_dev, &adcPar.t1, adcPar.t1.IrqStep * adcPar.t1.Pages,
                                            adcPar.t1.NCh, L_STREAM_ADC);
    }

    if (err.isSuccess()) {
        err = lcompDeviceStart();
    }

    if (err.isSuccess()) {
        m_adcReceiver.prestartInit();
        m_adcReceiver.start();
    }

    return err;
}

LQMeasError lcompStreamDevice::protAdcStop() {
    LQMeasError err = LQMeasError::Success();

    m_adcReceiver.stop();
    m_adcReceiver.wait();

    if (m_dev->StopLDevice()!=L_SUCCESS) {
        err = lcompErrors::AdcStop();
    }

    return err;
}

LQMeasError lcompStreamDevice::protAdcSetRawConfig(const void *rawCfg, size_t cfg_size) {
    LQMeasError err = LQMeasError::Success();
    if (cfg_size != sizeof(AdcConfig)) {
        err = LQMeasStdErrors::InvalidAdcConfig();
    } else {
        memcpy(&m_adcCfg, rawCfg, cfg_size);
        ADC_PAR adcPar;
        err = fillAdcParams(&adcPar);
    }
    return err;
}

LQMeasError lcompStreamDevice::protLoadConfig(QSettings &set) {
    AdcConfig cfg;
    LQMeasError err = LQMeasError::Success();

    unsigned ch_cnt = set.beginReadArray("Channels");
    if (ch_cnt == 0) {
        ch_cnt = 2;
        cfg.Ch[0].Enabled = true;
        cfg.Ch[0].Mode = AdcConfig::ADC_MODE_DIFF;
        cfg.Ch[0].Range = 0;

        cfg.Ch[1].Enabled = true;
        cfg.Ch[1].Mode = AdcConfig::ADC_MODE_DIFF;
        cfg.Ch[1].Range = 0;
    } else {
        for (unsigned ch=0; (ch < ch_cnt) && (ch< adcChannelsCnt()); ch++) {
            set.setArrayIndex(ch);
            cfg.Ch[ch].Enabled = set.value("Enabled").toBool();
            cfg.Ch[ch].Range   = set.value("Range").toInt();
            cfg.Ch[ch].Mode    = (AdcConfig::ADC_MODE)set.value("Mode").toInt();
        }
    }

    for (unsigned ch=ch_cnt; ch < adcChannelsCnt(); ch++) {
        cfg.Ch[ch].Enabled = false;
        cfg.Ch[ch].Range = 0;
        cfg.Ch[ch].Mode = AdcConfig::ADC_MODE_DIFF;
    }
    set.endArray();

    cfg.FreqAdc = set.value("FreqAdc", adcFreqMax()).toDouble();
    cfg.FreqCh  = set.value("FreqCh").toDouble();

    err = adcSetRawConfig(&cfg, sizeof(cfg));
    if (err.isSuccess())
        err = configure();
    return err;
}

LQMeasError lcompStreamDevice::protSaveConfig(QSettings &set) const {
    LQMeasError err = LQMeasError::Success();

    set.beginWriteArray("Channels", adcChannelsCnt());

    for (unsigned ch=0; ch< adcChannelsCnt(); ch++) {
        set.setArrayIndex(ch);
        set.setValue("Enabled", m_adcCfg.Ch[ch].Enabled);
        set.setValue("Range", m_adcCfg.Ch[ch].Range);
        set.setValue("Mode", m_adcCfg.Ch[ch].Mode);
    }
    set.endArray();

    set.setValue("FreqAdc", m_adcCfg.FreqAdc);
    set.setValue("FreqCh", m_adcCfg.FreqCh);
    return err;
}

LQMeasError lcompStreamDevice::fillAdcParams(void *par) {
    ADC_PAR *padcPar = (ADC_PAR *)par;
    unsigned nch = 0;
    LQMeasError err = LQMeasError::Success();

    padcPar->t1.s_Type = L_ADC_PARAM;
    padcPar->t1.AutoInit = 1;
    padcPar->t1.dScale = 0;
    padcPar->t1.SynchroType = 0; //3
    padcPar->t1.SynchroSensitivity = 0;
    padcPar->t1.SynchroMode = 0;
    padcPar->t1.AdChannel = 0;
    padcPar->t1.AdPorog = 0;

    for (unsigned ch=0; (ch < adcChannelsCnt()) && err.isSuccess(); ch++) {
        if (m_adcCfg.Ch[ch].Enabled) {
            ULONG lch = 0;
            lch |= (m_adcCfg.Ch[ch].Range << 6);
            if (m_adcCfg.Ch[ch].Mode == AdcConfig::ADC_MODE_ZERO) {
                lch |= (1<<4);
            } else if (m_adcCfg.Ch[ch].Mode == AdcConfig::ADC_MODE_COMM) {
                lch |= (1<<5) | ch;
            } else if (m_adcCfg.Ch[ch].Mode == AdcConfig::ADC_MODE_DIFF) {
                if (ch < 16) {
                    lch |= ch;
                } else {
                    err = LQMeasStdErrors::InvalidAdcConfig();
                }
            } else {
                err = LQMeasStdErrors::InvalidAdcConfig();
            }

            if (err.isSuccess()) {
                padcPar->t1.Chn[nch++] = lch;
            }
        }
    }

    padcPar->t1.dRate = m_adcCfg.FreqAdc/1000;
    padcPar->t1.dKadr = (m_adcCfg.FreqCh <= 0) || (m_adcCfg.FreqCh > m_adcCfg.FreqAdc/nch) ? 0 :
                          1000./m_adcCfg.FreqCh - 1000 *(nch-1)/m_adcCfg.FreqAdc;
    padcPar->t1.NCh = nch;

    unsigned irq_step = m_adcCfg.FreqAdc/100;
    if (irq_step == 0) {
        irq_step = 64;
    } else {
        irq_step += 63;
        irq_step -= irq_step % 64;
    }


    padcPar->t1.FIFO = 4096;
    padcPar->t1.IrqStep = irq_step;
    padcPar->t1.Pages = 256;

    padcPar->t1.IrqEna = 1;
    padcPar->t1.AdcEna = 1;

    if (m_dev->FillDAQparameters(&padcPar->t1)!=L_SUCCESS) {
        err =  lcompErrors::AdcFillParams();
    }

    if (err.isSuccess()) {
        m_adcCfg.FreqAdc = 1000. * padcPar->t1.dRate;
        m_adcCfg.FreqCh = 1000./(1000.*(nch-1)/ m_adcCfg.FreqAdc + padcPar->t1.dKadr);
    }
    return err;
}
