#include "ltr210.h"
#include "../../../lqtdefs.h"
#include "../crates/LtrCrate.h"

#include <QElapsedTimer>
#include <QScopedArrayPointer>

static double f_ranges[] = {10.,5.,2.,1.,0.5};



LTR210::LTR210(QSharedPointer<LtrCrate> crate, unsigned slot, QObject *parent) :
    LtrModule(crate, slot, typeModuleID, parent) {
    adcSetChCnt(typeAdcChCnt);
    LTR210_Init(&m_hnd);
}


QString LTR210::typeModuleName() {
    return "LTR210";
}

LTR210::~LTR210() {
    if (isOpened())
        close();
}

QString LTR210::name() const {
    return typeModuleName();
}

QString LTR210::serial() const {
    return QSTRING_FROM_CSTR(m_hnd.ModuleInfo.Serial);
}

unsigned LTR210::verFPGA() const {
    return m_hnd.ModuleInfo.VerFPGA;
}

unsigned LTR210::verPLD() const {
    return m_hnd.ModuleInfo.VerPLD;
}

double LTR210::adcFreq() const {
    return m_hnd.State.AdcFreq;;
}

double LTR210::adcChRange(int ch_num) const {
    Q_ASSERT((ch_num>=0) && (ch_num < adcChannelsCnt()));
    return f_ranges[m_hnd.Cfg.Ch[ch_num].Range];
}

const double LTR210::adcRangeMaxVal(int range) const {
    return f_ranges[range];
}

const double LTR210::adcRangeMinVal(int range) const {
    return -f_ranges[range];
}

unsigned LTR210::adcRangesCnt() const {
    return sizeof(f_ranges)/sizeof(f_ranges[0]);
}

double LTR210::adcFreqMax() const {
    return 10000000;
}

QString LTR210::errorString(int err) const {
    return QSTRING_FROM_CSTR(LTR210_GetErrorString(err));
}

bool LTR210::isOpened() const {
    return LTR210_IsOpened(&m_hnd) == LTR_OK;
}

bool LTR210::adcIsRunning() const {
    return m_hnd.State.Run ? true : false;
}

bool LTR210::adcChEnabled(int ch) const {
    Q_ASSERT((ch>=0) && (ch < adcChannelsCnt()));
    return m_hnd.Cfg.Ch[ch].Enabled;
}

LQMeasError LTR210::getFrame(double *data, unsigned size, unsigned flags, unsigned tout) {
    LQMeasError err = LQMeasError::Success();
    INT wrds_cnt = size;
    QElapsedTimer tmr;

    tmr.start();

    if (m_hnd.Cfg.SyncMode!=LTR210_SYNC_MODE_CONTINUOUS) {
        wrds_cnt = m_hnd.State.RecvFrameSize;
        DWORD evt=LTR210_RECV_EVENT_TIMEOUT;       

        while ((evt!=LTR210_RECV_EVENT_SOF) && !tmr.hasExpired(tout) && err.isSuccess()) {
            err = error(LTR210_WaitEvent(&m_hnd, &evt, NULL, (tout - tmr.elapsed())));
            if (!err.isSuccess()) {
                reportError(tr("Ошибка ожидания данных от модуля"), err);
            }
        }

        if (err.isSuccess() && (evt!=LTR210_RECV_EVENT_SOF)) {
            err = error(LTR_ERROR_RECV_INSUFFICIENT_DATA);
            reportError(tr("Не удалось дождаться приема начала кадра"), err);
        }
    }

    if (err.isSuccess()) {
        QScopedArrayPointer<DWORD> wrds(new DWORD[wrds_cnt]);
        QScopedArrayPointer<double> tmp_data(((INT)size<(wrds_cnt-1)) ?
                                             new double[wrds_cnt] : 0);
        if (err.isSuccess()) {
            INT rcv = LTR210_Recv(&m_hnd, wrds.data(), 0, wrds_cnt, (tout - tmr.elapsed()));

            if (rcv < 0) {
                err = error(rcv);
                reportError(tr("Ошибка приема данных"), err);
            } else if (rcv < wrds_cnt) {
                err = error(LTR_ERROR_RECV_INSUFFICIENT_DATA);
                reportError(tr("Принято недостаточно данных. Запрошено %1, принято %2")
                            .arg(QString::number(wrds_cnt))
                            .arg(QString::number(rcv)),
                            err);
            }
        }

        if (err.isSuccess()) {
            DWORD proc_flags = LTR210_PROC_FLAG_VOLT | LTR210_PROC_FLAG_AFC_COR | LTR210_PROC_FLAG_ZERO_OFFS_COR;
            TLTR210_FRAME_STATUS status;

            err = error(LTR210_ProcessData(&m_hnd, wrds.data(), tmp_data ? tmp_data.data() : data, &wrds_cnt,
                                            proc_flags, &status, NULL));
            if (!err.isSuccess()) {
                reportError("Ошибка обработки данных", err);
            } else if (status.Result==LTR210_FRAME_RESULT_ERROR) {
                err = error(LTR_ERROR_RECV);
                reportError("Неверные данные в принятом кадре!", err);
            }

            if (tmp_data) {
                memcpy(data, tmp_data.data(), size);
            }
        }
    }

    return err;
}

LQMeasError LTR210::waitSof(unsigned tout) {
    LQMeasError err = LQMeasError::Success();
    QElapsedTimer tmr;

    tmr.start();

    if (m_hnd.Cfg.SyncMode!=LTR210_SYNC_MODE_CONTINUOUS) {
        DWORD evt=LTR210_RECV_EVENT_TIMEOUT;

        while ((evt!=LTR210_RECV_EVENT_SOF) && !tmr.hasExpired(tout) && err.isSuccess()) {
            err = error(LTR210_WaitEvent(&m_hnd, &evt, NULL, (tout - tmr.elapsed())));
            if (!err.isSuccess()) {
                reportError(tr("Ошибка ожидания данных от модуля"), err);
            }
        }

        if (err.isSuccess() && (evt!=LTR210_RECV_EVENT_SOF)) {
            err = error(LTR_ERROR_RECV_INSUFFICIENT_DATA);
            reportError(tr("Не удалось дождаться приема начала кадра"), err);
        }
    }
    return err;
}

LQMeasError LTR210::adcGetData(double *data, unsigned size, unsigned flags, unsigned tout,
                               unsigned *recvd_size, TLTR210_FRAME_STATUS *status) {
    LQMeasError err = LQMeasError::Success();
    int wrds_cnt = size;
    int recvd_words=0;

    QScopedArrayPointer<DWORD> wrds(new DWORD[wrds_cnt]);

    if (err.isSuccess()) {
        INT rcv = LTR210_Recv(&m_hnd, wrds.data(), NULL, wrds_cnt, tout);
        if (rcv < 0) {
            err = error(rcv);
            reportError(tr("Ошибка приема данных"), err);
        } else {
            recvd_words = rcv;
        }
    }

    if (err.isSuccess() && (recvd_words!=0)) {
        DWORD proc_flags = LTR210_PROC_FLAG_VOLT | LTR210_PROC_FLAG_AFC_COR | LTR210_PROC_FLAG_ZERO_OFFS_COR;
        TLTR210_FRAME_STATUS cur_status;

        err = error(LTR210_ProcessData(&m_hnd, wrds.data(), data, &recvd_words,
                                        proc_flags, &cur_status, NULL));
        if (!err.isSuccess()) {
            reportError("Ошибка обработки данных", err);
        } else if (cur_status.Result==LTR210_FRAME_RESULT_ERROR) {
            reportError("Неверные данные в принятом кадре!", err);
        } else if (recvd_words > (int)adcEnabledChCnt()){
            for (unsigned ch=0; ch < adcChannelsCnt(); ch++) {
                if (adcChEnabled(ch))
                    adcSetChStatus(ch, ChStatusOk);
            }
        }

        if (status)
            *status = cur_status;
    }

    if (err.isSuccess() && recvd_size)
        *recvd_size = recvd_words;

    return err;
}


LQMeasError LTR210::startFrame() {
    LQMeasError err = LQMeasError::Success();
    if (!m_hnd.State.Run) {
        err = adcStart();
    }

    if (err.isSuccess() && (m_hnd.Cfg.SyncMode==LTR210_SYNC_MODE_INTERNAL)) {
        err = error(LTR210_FrameStart(&m_hnd));
        if (!err.isSuccess()) {
            reportError(tr("Ошибка программного запуска кадра"), err);
        }
    }
    return err;
}

LQMeasError LTR210::measZeroAllRanges() {
    TLTR210_CONFIG cfg = m_hnd.Cfg;
    LQMeasError err = LQMeasError::Success();

    reportMessage(QxtLogger::InfoLevel, tr("Запуск измерения нуля по всем диапазонам"));
    for (unsigned range = 0; (range < LTR210_RANGE_CNT) && err.isSuccess(); range++) {
        for (unsigned ch=0; ch < adcChannelsCnt(); ch++)
            m_hnd.Cfg.Ch[ch].Range = range;
        err = error(LTR210_MeasAdcZeroOffset(&m_hnd, 0));
        if (err.isSuccess()) {
            for (unsigned ch=0; ch < adcChannelsCnt(); ch++)
                m_zeroOffs[ch][range] = m_hnd.State.AdcZeroOffset[ch];
        } else {
            reportError(tr("Не удалось выполнить измерение смещения нуля"), err);
        }
    }

    LQMeasError setErr = adcSetRawConfig(&cfg, sizeof(cfg));
    if (setErr.isSuccess())
        setErr = configure();
    if (err.isSuccess())
        err = setErr;

    if (err.isSuccess())
        reportMessage(QxtLogger::InfoLevel, tr("Измерение нуля завершено успешно"));
    return err;
}


LQMeasError LTR210::loadFPGA(QString filename) {
    LQMeasError err;

    reportMessage(QxtLogger::InfoLevel, tr("Начата загрузка ПЛИС"));
    err = error(LTR210_LoadFPGA(&m_hnd, filename.toAscii(), cb_load_porgr, this));

    sendOpProgess(OP_FINISH, loadFpgaOpDescr(), 0, 0, err);

    if (!err.isSuccess()) {
        reportError(tr("Ошибка загрузки прошивки ПЛИС"), err);
    } else {
        reportMessage(QxtLogger::InfoLevel, tr("Загрузка ПЛИС завершена успешно"));
        emit devDescrChanged();
    }
    return err;
}

LQMeasError LTR210::protAdcSetRawConfig(const void *cfg, size_t cfg_size) {
    LQMeasError err = LQMeasError::Success();
    if (cfg_size != sizeof(TLTR210_CONFIG)) {
        err = LQMeasStdErrors::InvalidAdcConfig();
    } else {
        memcpy(&m_hnd.Cfg, cfg, cfg_size);
    }
    return err;
}

LQMeasError LTR210::adcGetData(double *data, unsigned size, unsigned flags,
                       unsigned tout, unsigned *recvd_size) {
    return adcGetData(data, size, flags, tout, recvd_size, 0);
}

TLTR *LTR210::channel()  const {
    return &m_hnd.Channel;
}

LQMeasError LTR210::protOpen() {
    LQMeasError err = error(LTR210_Open(&m_hnd, crate()->ltrdIpAddr(), crate()->ltrdTcpPort(),
                                crate()->serial().toAscii(), slot()));
    if (err.isSuccess() && (LTR210_FPGAIsLoaded(&m_hnd)!=LTR_OK))
        err = loadFPGA("");

    if (err.isSuccess()) {
        for (unsigned ch=0; ch < adcChannelsCnt(); ch++) {
            for (unsigned range=0; range < LTR210_RANGE_CNT; range++) {
                m_zeroOffs[ch][range] = 0;
            }
        }
    }

    if (!err.isSuccess())
        LTR210_Close(&m_hnd);

    return err;
}

LQMeasError LTR210::protClose() {
    return error(LTR210_Close(&m_hnd));
}

LQMeasError LTR210::protConfigure() {
    LQMeasError err = error(LTR210_SetADC(&m_hnd));

    if (err.isSuccess()) {
        for (unsigned ch=0; ch < adcChannelsCnt(); ch++) {
            m_hnd.State.AdcZeroOffset[ch] = m_zeroOffs[ch][m_hnd.Cfg.Ch[ch].Range];
        }
    }
    return err;
}

LQMeasError LTR210::protLoadConfig(QSettings &set) {
    /** @todo */
    return error(LTR_ERROR_NOT_IMPLEMENTED);
}

LQMeasError LTR210::protSaveConfig(QSettings &set) const {
    /** @todo */
    return error(LTR_ERROR_NOT_IMPLEMENTED);
}

LQMeasError LTR210::protAdcStart() {
    return error(LTR210_Start(&m_hnd));
}

LQMeasError LTR210::protAdcStop() {
    return error(LTR210_Stop(&m_hnd));
}

void LTR210::cb_load_porgr(void *cb, TLTR210 *hnd, DWORD done_cnt, DWORD total_cnt) {
    LTR210 *dev = reinterpret_cast<LTR210*>(cb);
    if (dev) {
        dev->sendOpProgess(done_cnt == 0 ? OP_START : OP_PROGR, loadFpgaOpDescr(),
                           done_cnt, total_cnt, LQMeasError::Success());
    }
}
