#include "ltr24.h"
#include "../../../lqtdefs.h"
#include "../crates/LtrCrate.h"

static double f_ranges[] = {2., 10.};

static const double f_adc_freqs[] = {
        117.1875e3,  78.125e3,   58.59375e3,   39.0625e3,
        29.296875e3, 19.53125e3, 14.6484375e3, 9.765625e3,
        7.32421875e3, 4.8828125e3, 3.662109375e3, 2.44140625e3,
        1.8310546875e3, 1.220703125e3, 0.91552734375e3, 0.6103515625e3
};

QString LTR24::typeModuleName() {
    return "LTR24";
}


LTR24::LTR24(QSharedPointer<LtrCrate> crate, unsigned slot, QObject *parent) :
    LtrModule(crate, slot, typeModuleID, parent), m_receiver(this) {
    LTR24_Init(&m_hnd);
    adcSetChCnt(LTR24_CHANNEL_NUM);
}


bool LTR24:: isOpened() const {
    return LTR24_IsOpened(&m_hnd)==LTR_OK;
}

bool LTR24::adcIsRunning() const {
    return m_hnd.Run ? true : false;
}

bool LTR24::adcChEnabled(int ch) const {
    Q_ASSERT((ch>=0) && (ch < LTR24_CHANNEL_NUM));
    return m_hnd.ChannelMode[ch].Enable;
}

LQMeasError LTR24::adcGetData(double *data, unsigned size, unsigned flags, unsigned tout,
                      unsigned *recvd_size) {
    LQMeasError err = LQMeasError::Success();
    int wrds_cnt = m_hnd.DataFmt == LTR24_FORMAT_20 ? size : 2*size;
    unsigned recvd_words=0;
    DWORD* wrds;
    QScopedArrayPointer<BOOL> ovrlds(new BOOL[size]);
    INT proc_size = 0;

    if (err.isSuccess()) {
        err = m_receiver.getFrames(wrds_cnt, tout, &wrds, &recvd_words);
    }

    if (err.isSuccess() && (recvd_words!=0)) {
        DWORD proc_flags = LTR24_PROC_FLAG_VOLT | LTR24_PROC_FLAG_AFC_COR | LTR24_PROC_FLAG_CALIBR;
        proc_size = (INT)recvd_words;

        err = error(LTR24_ProcessData(&m_hnd, wrds, data, &proc_size,
                                      proc_flags, ovrlds.data()));
        if (!err.isSuccess()) {
            reportError("Ошибка обработки данных", err);
        } else {
            if (m_hnd.DataFmt == LTR24_FORMAT_24) {
                for (unsigned ch_idx = 0; ch_idx < adcEnabledChCnt(); ch_idx++) {
                    ChannelStatus status = ChStatusOk;
                    for (unsigned i=ch_idx; ((INT)i < proc_size) && (status == ChStatusOk); i+=adcEnabledChCnt()) {
                        if (ovrlds[i]) {
                            status = ChStatusOverload;
                        }
                    }
                    adcSetChStatus(adcChNum(ch_idx), status);
                }
            }
        }
    }

    if (err.isSuccess() && recvd_size)
        *recvd_size = proc_size;

    return err;
}

TLTR *LTR24::channel() const {
    return &m_hnd.Channel;
}

LQMeasError LTR24::protOpen() {
    LQMeasError err = error(LTR24_Open(&m_hnd, crate()->ltrdIpAddr(), crate()->ltrdTcpPort(),
                            crate()->serial().toAscii(), slot()));
    if (err.isSuccess()) {
        err = error(LTR24_GetConfig(&m_hnd));
    }

    if (!err.isSuccess()) {
        LTR24_Close(&m_hnd);
    }
    return err;
}

LQMeasError LTR24::protClose() {
    return error(LTR24_Close(&m_hnd));
}

LQMeasError LTR24::protConfigure() {
    return error(LTR24_SetADC(&m_hnd));
}

LQMeasError LTR24::protLoadConfig(QSettings &set) {
    /** @todo */
    return error(LTR_ERROR_NOT_IMPLEMENTED);
}

LQMeasError LTR24::protSaveConfig(QSettings &set) const {
    /** @todo */
    return error(LTR_ERROR_NOT_IMPLEMENTED);
}

LQMeasError LTR24::protAdcSetRawConfig(const void *rawCfg, size_t cfg_size) {
    LQMeasError err = LQMeasError::Success();
    if (cfg_size!=sizeof(AdcConfig)) {
        err = error(LTR_ERROR_PARAMETERS);
    } else {
        AdcConfig *cfg = (AdcConfig*)rawCfg;

        m_hnd.ADCFreqCode = cfg->ADCFreqCode;
        m_hnd.DataFmt = cfg->DataFmt;
        m_hnd.ISrcValue = cfg->ISrcValue;
        m_hnd.TestMode = cfg->TestMode;
        for (int ch=0; ch < LTR24_CHANNEL_NUM; ch++) {
            m_hnd.ChannelMode[ch].Enable = cfg->ChannelMode[ch].Enable;
            m_hnd.ChannelMode[ch].AC = cfg->ChannelMode[ch].AC;
            m_hnd.ChannelMode[ch].Range = cfg->ChannelMode[ch].Range;
            m_hnd.ChannelMode[ch].ICPMode = cfg->ChannelMode[ch].ICPMode;
        }
    }
    return err;
}

LQMeasError LTR24::rawWordsReceive(DWORD *wrds, unsigned size, unsigned tout,
                            unsigned *recvd_size) {
    LQMeasError err = LQMeasError::Success();
    INT recv_res = LTR24_Recv(&m_hnd, wrds, 0, size, tout);
    if (recv_res >= 0) {
        *recvd_size = recv_res;
    } else {
        err = error(recv_res);
    }
    return err;
}

LQMeasError LTR24::protAdcStart() {
    LQMeasError err = error(LTR24_Start(&m_hnd));
    if (err.isSuccess()) {
        m_receiver.setFrameSize(adcEnabledChCnt() *
                                (m_hnd.DataFmt == LTR24_FORMAT_20 ? 1 : 2));
    }
    return err;
}

LQMeasError LTR24::protAdcStop() {
    return error(LTR24_Stop(&m_hnd));
}


QString LTR24::serial() const {
    return QSTRING_FROM_CSTR(m_hnd.ModuleInfo.Serial);
}

QString LTR24::name() const {
    return typeModuleName();
}

QString LTR24::modificationName() const {
    return QString("%1-%2").arg(name()).arg(m_hnd.ModuleInfo.SupportICP ? "2" : "1");
}

int LTR24::pldVer() const {
    return m_hnd.ModuleInfo.VerPLD;
}

bool LTR24::supportICP() const {
    return m_hnd.ModuleInfo.SupportICP;
}

const double LTR24::adcRangeMaxVal(int range) const {
    return f_ranges[range];
}

const double LTR24::adcRangeMinVal(int range) const {
    return -f_ranges[range];
}

unsigned LTR24::adcRangesCnt() const {
    return sizeof(f_ranges)/sizeof(f_ranges[0]);
}

double LTR24::adcFreqMax() const {
    return f_adc_freqs[0];
}

double LTR24::typeAdcFreqVal(int idx) {
    return f_adc_freqs[idx];
}

unsigned LTR24::typeAdcFreqsCnt() {
 return sizeof(f_adc_freqs)/sizeof(f_adc_freqs[0]);
}

double LTR24::adcFreq() const {
    return m_hnd.ADCFreq;
}

double LTR24::adcChRange(int ch) const {
    Q_ASSERT((ch>=0) && (ch < LTR24_CHANNEL_NUM));
    return f_ranges[m_hnd.ChannelMode[ch].Range];
}

QString LTR24::errorString(int err) const {
    return QSTRING_FROM_CSTR(LTR24_GetErrorString(err));
}
