#include "x502.h"
#include "lqmeas/lqtdefs.h"
#include "lqmeas/StdErrors.h"
#include "lqmeas/log/Log.h"

#include <QSettings>
#include <QMutexLocker>

static double f_ranges[] = {10.,5.,2.,1.,0.5, 0.2};

namespace LQMeas {
    X502::X502(const t_x502_devrec info) : DevAdc(this, DevAdc::TypeSequential),
        DevOutStream(this),
        m_opened(false), m_receiver(this), m_sender(this),
        m_inRunning(false), m_outRunning(false), m_outStreams(0),
        m_sync_freq_mode(X502_SYNC_INTERNAL), m_sync_start_mode(X502_SYNC_INTERNAL),
        m_ref_freq(X502_REF_FREQ_2000KHZ), m_pullups(0) {

        m_outFreq = outDacMaxFreq();
        m_devrec = info;
        m_hnd = X502_Create();
        memset(&m_devinfo, 0, sizeof(m_devinfo));
        adcSetChCnt(X502_ADC_COMM_CH_CNT);

        outSetRamSyncMode(OutRamSyncModeStream);

        AdcConfig cfg;
        memset(&cfg, 0, sizeof(cfg));
        cfg.FreqAdc = adcFreqMax();
        cfg.Ch[0].Enabled = true;
        m_adcCfg = cfg;
    }

    void X502::setNewDevinfo(const t_x502_devrec info) {
        X502_FreeDevRecordList(&m_devrec,1);
        m_devrec = info;
    }




    X502::~X502() {
        X502_Free(m_hnd);
        X502_FreeDevRecordList(&m_devrec,1);
    }

    QString X502::name() const {
        return QSTRING_FROM_CSTR(m_devrec.devname);
    }

    QString X502::serial() const {
        return QSTRING_FROM_CSTR(m_devinfo.serial);
    }

    QString X502::location() const {
        return QString::fromUtf8(m_devrec.location);
    }



    Device::DeviceInterface X502::iface() const {
        return m_devrec.iface == X502_IFACE_PCI ? Device::IFACE_PCI :
               m_devrec.iface == X502_IFACE_USB ? Device::IFACE_USB :
               m_devrec.iface == X502_IFACE_ETH ? Device::IFACE_ETH :
                                                  Device::IFACE_UNKNOWN;
    }


    bool X502::isOpened() const {
        return m_opened;
    }

    double X502::adcFreq() const {
        double freq=0;
        X502_GetAdcFreq(m_hnd, &freq, NULL);
        return freq;
    }

    double X502::adcChFreq() const {
        double chFreq=0;
        X502_GetAdcFreq(m_hnd, NULL, &chFreq);
        return chFreq;
    }

    double X502::adcChRange(int ch) const {
        double range = 0;
        Q_ASSERT((ch >= 0) && (ch<X502_ADC_COMM_CH_CNT));
        if (m_adcCfg.Ch[ch].Range < sizeof(f_ranges)/sizeof(f_ranges[0])) {
            range = f_ranges[m_adcCfg.Ch[ch].Range];
        }
        return range;
    }



    unsigned X502::adcRangesCnt() const {
        return sizeof(f_ranges)/sizeof(f_ranges[0]);
    }

    double X502::adcFreqMax() const {
        return 2000000;
    }

    const double X502::adcRangeMaxVal(int range) const {
        return f_ranges[range];
    }

    const double X502::adcRangeMinVal(int range) const {
        return -f_ranges[range];
    }

    bool X502::adcExternalStart() const {
        return (m_sync_start_mode != X502_SYNC_INTERNAL) || (m_sync_freq_mode != X502_SYNC_INTERNAL);
    }

    bool X502::adcIsRunning() const {
        return X502_IsRunning(m_hnd) == X502_ERR_OK ? true : false;
    }

    bool X502::adcChEnabled(int ch) const {
        return ch < X502_ADC_COMM_CH_CNT ?  m_adcCfg.Ch[ch].Enabled : false;
    }

    Error X502::adcEnable() {
        return error(X502_StreamsEnable(m_hnd, X502_STREAM_ADC));
    }

    Error X502::adcDisable() {
        return error(X502_StreamsDisable(m_hnd, X502_STREAM_ADC));
    }

    Error X502::protAdcSetRawConfig(const void *rawCfg, size_t cfg_size) {
        Error err = Error::Success();
        if (cfg_size != sizeof(AdcConfig)) {
            err = StdErrors::InvalidAdcConfig();
        } else {
            const AdcConfig* cfg = (const AdcConfig *)rawCfg;
            unsigned lch_cnt=0;

            double f_adc, f_frame;
            for (unsigned ch=0; (ch < X502_ADC_COMM_CH_CNT) && err.isSuccess(); ch++) {
                if (cfg->Ch[ch].Enabled) {
                    err = error(X502_SetLChannel(m_hnd, lch_cnt++, ch, cfg->Ch[ch].Mode, cfg->Ch[ch].Range, cfg->Ch[ch].Avg));
                }
            }

            if (err.isSuccess()) {
                err = error(X502_SetLChannelCount(m_hnd, lch_cnt));
                if (err.isSuccess())
                    m_lch_cnt = lch_cnt;
            }

            if (err.isSuccess()) {
                f_adc = cfg->FreqAdc;
                f_frame = cfg->FreqCh;
                err = error(X502_SetAdcFreq(m_hnd, &f_adc, &f_frame));
            }


            if (err.isSuccess()) {
                m_adcCfg = *cfg;
            }
        }
        return err;
    }


    Error X502::adcSetConfig(const X502::AdcConfig *cfg) {
        return adcSetRawConfig(cfg, sizeof(*cfg));
    }

    Error X502::adcGetData(double *data, unsigned size, unsigned flags, unsigned tout, unsigned *recvd_size) {
        Error err = Error::Success();
        unsigned recvd_words=0;

        uint32_t *wrds;

        err = m_receiver.getFrames(size, tout, &wrds, &recvd_words);


        if ((err.isSuccess()) && (recvd_words!=0)) {
            uint32_t proc_flags = X502_PROC_FLAGS_VOLT;
            uint32_t proc_wrds = recvd_words;


            err = error(X502_ProcessAdcData(m_hnd, wrds, data, &proc_wrds, proc_flags));
            if (!err.isSuccess()) {
                LQMeasLog->error(tr("Process data error"), err, this);
            }

            recvd_words = proc_wrds;
        }

        if (err.isSuccess() && recvd_size)
            *recvd_size = recvd_words;

        return err;
    }

    X502::AdcConfig X502::adcCurConfig() const {
        return m_adcCfg;
    }


    Error X502::error(int err) {
        return err == X502_ERR_OK ? Error::Success() :
                                    Error(err, QSTRING_FROM_CSTR(X502_GetErrorString(err)),
                                                Error::ERRTYPE_DEVNATIVE_X502);
    }


    Error X502::protOpen() {
        Error err = error(X502_OpenByDevRecord(m_hnd, &m_devrec));
        if (err.isSuccess()) {
            err = error(X502_GetDevInfo(m_hnd, &m_devinfo));
        }
        if (err.isSuccess()) {
            m_opened = true;
            m_outStreams = 0;
            m_outRunning = false;
            m_inRunning = false;
            m_sync_freq_mode = X502_SYNC_INTERNAL;
            m_sync_start_mode = X502_SYNC_INTERNAL;
            m_ref_freq = X502_REF_FREQ_2000KHZ;
            m_pullups = 0;
            m_outFreq = outDacMaxFreq();

            outSetDacChCnt(m_devinfo.devflags & X502_DEVFLAGS_DAC_PRESENT ? X502_DAC_CH_CNT : 0);
            outSetDigChCnt(X502_DOUT_LINES_CNT);
            err = adcSetConfig(&m_adcCfg);
        }
        return err;
    }

    Error X502::protClose() {
        Error err = error(X502_Close(m_hnd));
        m_opened = false;
        return err;
    }

    Error X502::protConfigure() {
        return error(X502_Configure(m_hnd, 0));
    }

    Error X502::protLoadConfig(QSettings &set) {
        Error err = Error::Success();
        AdcConfig cfg;


        unsigned ch_cnt = set.beginReadArray("Channels");
        if (ch_cnt == 0) {
            ch_cnt = 1;
            cfg.Ch[0].Enabled = true;
            cfg.Ch[0].Mode = X502_LCH_MODE_DIFF;
            cfg.Ch[0].Range = X502_ADC_RANGE_10;
            cfg.Ch[0].Avg = 0;
        } else {

            for (unsigned ch=0; (ch < ch_cnt) && (ch< adcChannelsCnt()); ch++) {
                set.setArrayIndex(ch);
                cfg.Ch[ch].Enabled = set.value("Enabled").toBool();
                cfg.Ch[ch].Range   = (t_x502_adc_range)set.value("Range").toInt();
                cfg.Ch[ch].Mode    = (t_x502_lch_mode)set.value("Mode").toInt();
                cfg.Ch[ch].Avg     = set.value("Avg", 0).toInt();
            }
        }

        for (unsigned ch=ch_cnt; ch < adcChannelsCnt(); ch++) {
            cfg.Ch[ch].Enabled = false;
            cfg.Ch[ch].Range = X502_ADC_RANGE_10;
            cfg.Ch[ch].Mode = X502_LCH_MODE_COMM;
            cfg.Ch[ch].Avg = 0;
        }
        set.endArray();


        cfg.FreqAdc = set.value("FreqAdc", "2000000").toDouble();
        cfg.FreqCh  = set.value("FreqCh").toDouble();


        setSyncFreqMode((t_x502_sync_mode)set.value("SyncFreqMode", X502_SYNC_INTERNAL).toInt());
        setSyncStartMode((t_x502_sync_mode)set.value("SyncStartMode", X502_SYNC_INTERNAL).toInt());
        setRefFreq((t_x502_ref_freq)set.value("RefFreq", X502_REF_FREQ_2000KHZ).toInt());
        setPullups(set.value("Pullups", 0).toInt());

        outLoadSignalConfig(set);

        setOutFreq(set.value("OutFreq", m_outFreq).toDouble());

        DevOut::OutRamSyncMode outMode = (DevOut::OutRamSyncMode)set.value("OutMode", DevOut::OutRamSyncModeStream).toInt();
        if ((outMode == DevOut::OutRamSyncModeStream) || (outMode == DevOut::OutRamSyncModeCycle))
            outSetRamSyncMode(outMode);

        err = adcSetConfig(&cfg);
        return err;
    }

    Error X502::protSaveConfig(QSettings &set) const {
        set.beginWriteArray("Channels", adcChannelsCnt());

        for (unsigned ch=0; ch< adcChannelsCnt(); ch++) {
            set.setArrayIndex(ch);
            set.setValue("Enabled", m_adcCfg.Ch[ch].Enabled);
            set.setValue("Range", m_adcCfg.Ch[ch].Range);
            set.setValue("Mode", m_adcCfg.Ch[ch].Mode);
            set.setValue("Avg", m_adcCfg.Ch[ch].Avg);
        }
        set.endArray();

        set.setValue("FreqAdc", m_adcCfg.FreqAdc);
        set.setValue("FreqCh", m_adcCfg.FreqCh);


        set.setValue("SyncFreqMode", m_sync_freq_mode);
        set.setValue("SyncStartMode", m_sync_start_mode);
        set.setValue("RefFreq", m_ref_freq);
        set.setValue("Pullups", (int)m_pullups);

        set.setValue("OutFreq", m_outFreq);
        set.setValue("OutMode", (int)outRamSyncMode());

        outSaveSignalConfig(set);

        return Error::Success();
    }

    Error X502::protAdcStart() {
        Error err = Error::Success();
        m_receiver.setFrameSize(adcEnabledChCnt());

        QMutexLocker lock(&m_streamLock);
        if (!m_outRunning)
            err = error(X502_StreamsStart(m_hnd));
        if (err.isSuccess())
            m_inRunning = true;
        return err;
    }

    Error X502::protAdcStop() {
        QMutexLocker lock(&m_streamLock);
        m_inRunning = false;
        return m_outRunning == 0 ? error(X502_StreamsStop(m_hnd)) : Error::Success();
    }

    Error X502::rawWordsReceive(uint32_t *wrds, unsigned size, unsigned tout, unsigned *recvd_size) {
        Error err = Error::Success();
        int32_t recv_res = X502_Recv(m_hnd, wrds, size, tout);
        if (recv_res >= 0) {
            *recvd_size = recv_res;
        } else {
            err = error(recv_res);
        }
        return err;
    }






    double X502::outDacMaxVal(int ch) const {
        return X502_DAC_RANGE;
    }

    double X502::outDacMinVal(int ch) const {
        return -X502_DAC_RANGE;
    }

    double X502::outDacMaxFreq() const {
        return X502_REF_FREQ_2000KHZ/X502_OUT_FREQ_DIV_MIN;
    }

    Error X502::setOutFreq(double val) {
        Error err = error(X502_SetOutFreq(m_hnd, &val));
        if (err.isSuccess())
            m_outFreq = val;
        return err;
    }

    Error X502::setSyncFreqMode(t_x502_sync_mode mode) {
        Error err = error(X502_SetSyncMode(m_hnd, mode));
        if (err.isSuccess())
            m_sync_freq_mode = mode;
        return err;
    }

    Error X502::setSyncStartMode(t_x502_sync_mode mode) {
        Error err = error(X502_SetSyncStartMode(m_hnd, mode));
        if (err.isSuccess())
            m_sync_start_mode = mode;
        return err;
    }

    Error X502::setRefFreq(t_x502_ref_freq freq) {
        Error err = error(X502_SetRefFreq(m_hnd, freq));
        if (err.isSuccess())
            m_ref_freq = freq;
        return err;
    }

    Error X502::setPullups(uint32_t pullups)     {
        Error err = error(X502_SetDigInPullup(m_hnd, pullups));
        if (err.isSuccess())
            m_pullups = pullups;
        return err;
    }

    Error X502::setOutStreams()  {
        int32_t err_code = X502_ERR_OK;
        uint32_t new_streams = 0;
        if (outDacChannelsCnt() >= X502_DAC_CH_CNT) {
            if (outDacValidRamSignal(0))
                new_streams |= X502_STREAM_DAC1;
            if (outDacValidRamSignal(1))
                new_streams |= X502_STREAM_DAC2;
        }
        bool outDig = false;
        for (int ch=0; (ch < outDigChannelsCnt()) && !outDig; ch++) {
            if (outDigValidRamSignal(ch))
                outDig = true;
        }
        if (outDig)
            new_streams |= X502_STREAM_DOUT;

        uint32_t streams_en = new_streams & ~m_outStreams;
        uint32_t streams_dis = ~new_streams & m_outStreams;

        if (streams_en) {
            err_code = X502_StreamsEnable(m_hnd, streams_en);
            if (err_code == X502_ERR_OK)
                m_outStreams |= streams_en;
        }
        if (streams_dis && (err_code == X502_ERR_OK)) {
            err_code = X502_StreamsDisable(m_hnd, streams_dis);
            if (err_code == X502_ERR_OK)
                m_outStreams &= ~streams_dis;
        }
        return error(err_code);
    }

    unsigned X502::curOutRamChannels() {
        return outDacEnabledChCnt() + (outDigEnabledChCnt() != 0 ? 1 : 0);
    }

    Error X502::protOutCycleLoadStart(unsigned size) {
        Error err = setOutStreams();
        if (err.isSuccess()) {
            err = error(X502_OutCycleLoadStart(m_hnd, size * curOutRamChannels()));
        }
        return err;
    }

    Error X502::protOutCycleLoadFinish() {
        return error(X502_OutCycleSetup(m_hnd, !m_outRunning ? X502_OUT_CYCLE_FLAGS_WAIT_DONE : 0));
    }

    Error X502::protOutCycleGenStart() {
        Error err = Error::Success();
        QMutexLocker lock(&m_streamLock);
        if (!m_inRunning && !m_outRunning)
            err = error(X502_StreamsStart(m_hnd));
        if (err.isSuccess())
            m_outRunning = true;
        return err;
    }

    Error X502::protOutCycleGenStop() {
        QMutexLocker lock(&m_streamLock);
        m_outRunning = false;
        /* деалаем корректную остановку, дожидаясь конца цикла (чтобы не остались неизвестные значения) */
        Error err = error(X502_OutCycleStop(m_hnd,X502_OUT_CYCLE_FLAGS_WAIT_DONE));
        if (m_inRunning == 0) {
            /* останавливаем поток только если поток ввода тоже уже остановлен */
            Error stop_err = error(X502_StreamsStop(m_hnd));
            if (err.isSuccess())
                err = stop_err;
        }
        return err;
    }

    Error X502::protOutStreamInit() {
        return setOutStreams();
    }

    Error X502::protOutStreamStart() {
        Error err = Error::Success();
        QMutexLocker lock(&m_streamLock);
        if (!m_inRunning)
            err = error(X502_StreamsStart(m_hnd));
        if (err.isSuccess())
            m_outRunning = true;
        return err;
    }

    Error X502::protOutStreamStop(unsigned tout) {
        QMutexLocker lock(&m_streamLock);
        m_outRunning = false;
        return m_inRunning == 0 ? error(X502_StreamsStop(m_hnd)) : Error::Success();
    }

    Error X502::rawWordsSend(const uint32_t *wrds, unsigned size, unsigned tout, unsigned *sent_size) {
        Error err = Error::Success();
        int send_res = X502_Send(m_hnd, wrds, size, tout);
        if (send_res >= 0) {
            *sent_size = send_res;
        } else {
            err = error(send_res);
        }
        return err;
    }

    Error X502::protOutSendData(const double *dac_data, unsigned dac_size,
                                         const unsigned *dig_data, unsigned dig_size,
                                         unsigned flags, unsigned tout) {
        Error err = Error::Success();

        if (m_sender.unsentWordsCnt()) {
            err = m_sender.flushData(tout);
        }

        if (err.isSuccess()) {
            if (m_sender.unsentWordsCnt()!=0) {
                err = StdErrors::SendBusy();
            } else {
                uint32_t snd_dac_req_size = dac_size;
                uint32_t snd_dig_req_size = dig_size;
                uint32_t snd_size = snd_dac_req_size  + snd_dig_req_size;
                QScopedArrayPointer<uint32_t> wrds (new uint32_t[snd_size]);

                if ((m_outStreams & X502_STREAM_DAC1) && (m_outStreams & X502_STREAM_DAC2)) {
                    dac_size/=2;
                    QScopedArrayPointer<double> dac1 (new double[dac_size]);
                    QScopedArrayPointer<double> dac2 (new double[dac_size]);
                    for (unsigned i=0; i < dac_size; i++) {
                        dac1[i] = *dac_data++;
                        dac2[i] = *dac_data++;
                    }
                    err = error(X502_PrepareData(m_hnd, dac1.data(), dac2.data(), (const uint32_t*)dig_data,
                                                 dac_size, X502_DAC_FLAGS_CALIBR  | X502_DAC_FLAGS_VOLT,
                                                 wrds.data()));
                } else {
                    err = error(X502_PrepareData(m_hnd, dac_data, dac_data, (const uint32_t*)dig_data,
                                                 dac_size, X502_DAC_FLAGS_CALIBR  | X502_DAC_FLAGS_VOLT,
                                                 wrds.data()));
                }

                if (!err.isSuccess()) {
                    LQMeasLog->error(tr("Prepare data error"), err, this);
                } else {
                    m_sender.setFrameSize(snd_size);
                    err = m_sender.putFrames(wrds.data(), snd_size, tout, 0);
                }
            }
        }
        return err;
    }

    Error X502::protOutFlushData(unsigned tout) {
        Error err = Error::Success();

        if (m_sender.unsentWordsCnt()) {
            err = m_sender.flushData(tout);
        }

        if (err.isSuccess()) {
            if (m_sender.unsentWordsCnt()!=0) {
                err = StdErrors::SendBusy();
            }
        }

        if (err.isSuccess()) {
            m_sender.setFrameSize(0);
        }
        return err;
    }
}
