#ifndef LQMEAS_X502_H
#define LQMEAS_X502_H

#include "x502api.h"
#include "lqmeas/devs/Device.h"
#include "lqmeas/devs/DeviceFrameReceiver.h"
#include "lqmeas/devs/DeviceFrameSender.h"
#include "lqmeas/ifaces/in/DevAdc.h"
#include "lqmeas/ifaces/out/DevOutStream.h"


#include <QMutex>

namespace LQMeas {
    class X502 : public Device, public DevAdc, public DevOutStream {
        Q_OBJECT
    public:

        struct AdcConfig {
            struct {
                bool Enabled;
                t_x502_lch_mode Mode;
                t_x502_adc_range Range;
                uint8_t Avg;
            } Ch[X502_ADC_COMM_CH_CNT];
            double FreqAdc;
            double FreqCh;
        } ;


        ~X502();


        QString name() const;
        QString serial() const;
        QString location() const;
        DeviceInterface iface() const;

        int pldaVer() const {return m_devinfo.plda_ver;}
        int fpgaVer() const {return m_devinfo.fpga_ver;}
        int mcuVer() const {return m_devinfo.mcu_firmware_ver;}

        bool galPresent() const {return m_devinfo.devflags & X502_DEVFLAGS_GAL_PRESENT ? true : false;}
        bool dacPresent() const {return m_devinfo.devflags & X502_DEVFLAGS_DAC_PRESENT ? true : false;}
        bool bfPresent() const {return m_devinfo.devflags & X502_DEVFLAGS_BF_PRESENT ? true : false;}
        bool ethPresent() const {return m_devinfo.devflags & X502_DEVFLAGS_IFACE_SUPPORT_ETH ? true : false;}

        bool isOpened() const;


        /* --------------------------- Adc Interface -----------------------------*/

        DevAdc *devAdc() {return static_cast<DevAdc*>(this);}

        double adcFreq() const;
        double adcChFreq() const;
        double adcChRange(int ch) const;
        const double *adcRanges() const;
        unsigned adcRangesCnt() const;
        double adcFreqMax() const;

        const double adcRangeMaxVal(int range) const;
        const double adcRangeMinVal(int range) const;


        bool adcExternalStart() const;
        bool adcIsRunning() const;
        bool adcChEnabled(int ch) const;


        Error adcEnable();
        Error adcDisable();


        Error adcSetConfig(const AdcConfig *cfg);
        Error adcGetData(double *data, unsigned size, unsigned flags,
                               unsigned tout, unsigned *recvd_size);

        AdcConfig adcCurConfig() const;



        /* ------------------- интерфейс вывода --------------------------------*/
        DevOutStream *devOutStream() {return static_cast<DevOutStream*>(this);}
        DevOut *devOut() {return static_cast<DevOut*>(this);}

        double outDacMaxVal(int ch) const;
        double outDacMinVal(int ch) const;
        double outDacMaxFreq() const;

        double outFreq() const {return m_outFreq;}
        bool outGenRunning() const {return m_outRunning;}

        Error setOutFreq(double val);

        OutChMode outDacChMode(int ch) const {return OutChModeRam;}
        OutChMode outDigChMode(int ch) const {return OutChModeRam;}


        /* ------------------- дополнительные функции --------------------------*/
        Error setSyncFreqMode(t_x502_sync_mode mode);
        Error setSyncStartMode(t_x502_sync_mode mode);
        Error setRefFreq(t_x502_ref_freq freq);
        Error setPullups(uint32_t pullups);


        t_x502_sync_mode syncFreqMode() const {return m_sync_freq_mode;}
        t_x502_sync_mode syncStartMode() const {return m_sync_start_mode;}
        t_x502_ref_freq refFreq() const {return m_ref_freq; }
        uint32_t pullups() const {return m_pullups;}




        static Error error(int err);

        t_x502_hnd devHandle() const {return m_hnd;}
    protected:
        X502(const t_x502_devrec info);



        Error protOpen();
        Error protClose();
        Error protConfigure();
        Error protLoadConfig(QSettings &set);
        Error protSaveConfig(QSettings &set) const;

        Error protAdcStart();
        Error protAdcStop();
        Error protAdcSetRawConfig(const void* cfg, size_t cfg_size);

        Error protOutCycleLoadStart(unsigned size);
        Error protOutCycleLoadFinish();
        Error protOutCycleGenStart();
        Error protOutCycleGenStop();

        Error protOutStreamInit();
        Error protOutStreamStart();
        Error protOutStreamStop(unsigned tout);

        Error rawWordsSend(const uint32_t *wrds, unsigned size, unsigned tout, unsigned *sent_size);
        Error protOutSendData(const double *dac_data, unsigned dac_size,
                               const unsigned *dig_data, unsigned dig_size, unsigned flags, unsigned tout);
        Error protOutFlushData(unsigned tout);
        unsigned curOutRamChannels();

        t_x502_info m_devinfo;
    private:
        Error rawWordsReceive(uint32_t *wrds, unsigned size, unsigned tout, unsigned *recvd_size);
        Error setOutStreams();
        void setNewDevinfo(const t_x502_devrec info);



        DeviceFrameReceiver<X502, uint32_t> m_receiver;
        DeviceFrameSender<X502, uint32_t> m_sender;

        t_x502_devrec m_devrec;
        t_x502_hnd m_hnd;
        bool m_opened;
        unsigned m_lch_cnt;

        AdcConfig m_adcCfg;
        double m_outFreq;
        uint32_t m_outStreams;
        bool     m_outRunning;
        uint32_t m_inRunning;
        t_x502_sync_mode m_sync_freq_mode;
        t_x502_sync_mode m_sync_start_mode;
        t_x502_ref_freq  m_ref_freq;
        uint32_t m_pullups;

        QMutex m_streamLock;

        friend class DeviceFrameReceiver<X502, uint32_t>;
        friend class DeviceFrameSender<X502, uint32_t>;
        friend class x502Resolver;

    };
}

#endif // LQMEAS_X502_H
