#include "DevAdc.h"
#include "lqmeas/devs/Device.h"
#include "lqmeas/log/Log.h"

namespace LQMeas {
    DevAdc::DevAdc(Device *dev, AdcType type) : m_type(type), m_run(false), m_dev(dev) {

    }

    Error DevAdc::adcSetRawConfig(const void *rawCfg, size_t cfg_size) {
        Error err = protAdcSetRawConfig(rawCfg, cfg_size);
        if (err.isSuccess()) {
            m_ch_indexes.clear();
            for (unsigned ch =0; ch < adcChannelsCnt(); ch++) {
                if (adcChEnabled(ch)) {
                    m_ch_indexes.append(ch);
                }
            }
        }
        return err;
    }


    unsigned DevAdc::adcChannelsCnt() const {
        return m_ch_status.size();
    }

    double DevAdc::adcChFreq() const {

        double ch_freq  = adcFreq();
        if (m_type == TypeSequential) {
            unsigned enChCnt = adcEnabledChCnt();
            if (enChCnt) {
                ch_freq/=enChCnt;
            }
        }
        return ch_freq;
    }

    Error DevAdc::adcStart() {
        Error err;

        for (int ch=0; ch < m_ch_status.size(); ch++) {
            m_ch_status[ch] = adcChEnabled(ch) ? DevAdc::ChStatusNoData :
                                                 DevAdc::ChStatusDisabled;
        }
        err = adcEnable();
        if (err.isSuccess())
            err = protAdcStart();

        if (err.isSuccess()) {
            m_run = true;
            LQMeasLog->detail(Device::tr("Data acquisition started successfully"), m_dev);
        } else {
            LQMeasLog->error(Device::tr("Data acquisition start error"), err, m_dev);
        }
        return err;
    }

    Error DevAdc::adcStop() {
        Error err, stop_err;
        err = adcDisable();
        stop_err = protAdcStop();
        if (err.isSuccess())
            err = stop_err;

        if (err.isSuccess()) {
            m_run = false;
            LQMeasLog->detail(Device::tr("Data acquisition stoped successfully"), m_dev);
        } else {
            LQMeasLog->error(Device::tr("Data acquisition stop error"), err, m_dev);
        }
        return err;
    }

    int DevAdc::adcChNum(int idx) const {
        return idx < m_ch_indexes.size() ? m_ch_indexes[idx] : -1;
    }

    int DevAdc::adcChIdx(int ch) const {
        return m_ch_indexes.indexOf(ch);
    }

    DevAdc::ChannelStatus DevAdc::adcGetChStatus(int ch, bool clear) {
        DevAdc::ChannelStatus ret;
        if (ch >= m_ch_status.size()) {
            ret = DevAdc::ChStatusError;
        } else {
            ret = m_ch_status[ch];
            if (clear && (ret!=DevAdc::ChStatusDisabled)) {
                m_ch_status[ch] = DevAdc::ChStatusNoData;
            }
        }
        return ret;
    }

    void DevAdc::adcSetChCnt(int ch_cnt) {
        m_ch_status.resize(ch_cnt);
    }

    void DevAdc::adcSetChStatus(int ch, DevAdc::ChannelStatus status) {
        if (ch < m_ch_status.size()) {
            /* состояние ошибочной ситуации по каналу должно сохраняться до сброса статуса */
            if ((m_ch_status[ch]==ChStatusNoData) || (status != ChStatusOk))
                m_ch_status[ch] = status;
        }
    }

    unsigned DevAdc::adcEnabledChCnt() const {
        return m_ch_indexes.size();
    }



    Error DevAdc::adcEnable() {
        return Error::Success();
    }

    Error DevAdc::adcDisable() {
        return Error::Success();
    }
}
