#include "DevOut.h"
#include <qmath.h>
#include <QtDebug>
#include "lqmeas/devs/Device.h"
#include "lqmeas/log/Log.h"
#include "lqmeas/ifaces/out/OutSignalFabric.h"


// функция вычисляет НОД
static int gcd( int a, int b ) {
    return b ? gcd( b, a % b ) : a;
}

// функция вычисляет НОК, используя НОД
static int lcm( int a, int b ){
    return a / gcd( a, b ) * b;
}


// source:
// http://www.isi.edu/~johnh/BLOG/1999/0728_RATIONAL_PI/fast_convergence.txt
// theory:
// https://en.wikipedia.org/wiki/Continued_fraction#Best_rational_approximations

static int rat_approx(double val, long* out_n, long* out_d, long max_d, double rel_err) {
    const size_t max_iter = 20;
    long coeff[max_iter];
    int sign = 1;
    double x = val;
    for (int i = 0; i < max_iter; ++i) {
        if (x < 0.0) {
            sign = -sign;
            x = -x;
        }
        long ipart = (long)x + (x - floor(x) > 0.5);
        coeff[i] = ipart * sign;
        x = 1.0 / (x - ipart);

        // compute numerator and denominator
        long num = coeff[i], den = 1;
        for (int j = i - 1; j >= 0; --j) {
            long t = num;
            num = coeff[j] * num + den;
            den = t;
        }
        if (den < 0) {
            num = -num;
            den = -den;
        }
        if (den > max_d)
            break;
        *out_n = num;
        *out_d = den;
        double approx_val = (double)num / den;
        double err = fabs(approx_val / val - 1.0);
        if (err <= rel_err)
            return 1;
    }
    return 0;
}


namespace LQMeas {
    DevOut::DevOut(Device *dev) : m_dev(dev), m_outRamSyncMode(OutRamSyncModeCycle),
        m_cycle_freq_rel_err(0.000001) {

    }

    int DevOut::outDacChannelsCnt() const {
        return m_sigsAnalog.size();
    }

    Error DevOut::outDacSetSignal(int ch, QSharedPointer<DacSignal> signal) {
        Q_ASSERT(ch < outDacChannelsCnt());
        m_sigsAnalog[ch] = signal;
        return Error::Success();
    }

    QSharedPointer<DacSignal> DevOut::outDacSignal(int ch) const {
        return (int)ch < m_sigsAnalog.size() ? m_sigsAnalog[ch] : QSharedPointer<DacSignal>();
    }

    Error DevOut::outDigSetSignal(int ch, QSharedPointer<DigSignal> signal) {
        Q_ASSERT(ch < outDigChannelsCnt());
        m_sigsDig[ch] = signal;
        return Error::Success();
    }

    QSharedPointer<DigSignal> DevOut::outDigSignal(int ch) const {
        return (int)ch < m_sigsDig.size() ? m_sigsDig[ch] : QSharedPointer<DigSignal>();
    }

    void DevOut::outGenRamInit() {
        for (int ch=0; ch < outDacChannelsCnt(); ch++) {
            if (outDacValidRamSignal(ch))
                outDacSignal(ch)->signalGenInit();
        }
        for (int ch=0; ch < outDigChannelsCnt(); ch++) {
            if (outDigValidRamSignal(ch))
                outDigSignal(ch)->signalGenInit();
        }
    }

    Error DevOut::outGenRamData(unsigned points, QVector<double> &dac_data,
                                QVector<unsigned> &dig_data) {
        QVector<int> dac_chs;
        QVector<int> dig_chs;
        for (int ch=0; ch < outDacChannelsCnt(); ch++) {
            if (outDacValidRamSignal(ch))
                dac_chs.append(ch);
        }
        for (int ch=0; ch < outDigChannelsCnt(); ch++) {
            if (outDigValidRamSignal(ch))
                dig_chs.append(ch);
        }

        if (dac_chs.size()) {
            dac_data.resize(points*dac_chs.size());
            for (int ch_idx=0; ch_idx < dac_chs.size(); ch_idx++) {
                QSharedPointer<DacSignal> sig = outDacSignal(dac_chs[ch_idx]);
                for (unsigned pt=0; pt < points; pt++) {
                    dac_data[pt*dac_chs.size()+ch_idx] = sig->signalGenNextPoint();
                }
            }
        } else {
            dac_data.clear();
        }

        if (dig_chs.size()) {
            dig_data.resize(points);
            memset(dig_data.data(), 0, dig_data.size()*sizeof(dig_data[0]));
            for (int ch_idx=0; ch_idx < dig_chs.size(); ch_idx++) {
                QSharedPointer<DigSignal> sig = outDigSignal(dig_chs[ch_idx]);

                for (unsigned pt=0; pt < points; pt++) {
                    if (sig->signalGenNextPoint()) {
                        dig_data[pt] |= (1 << dig_chs[ch_idx]);
                    }
                }
            }
        } else {
            dig_data.clear();
        }

        return Error::Success();
    }

    bool DevOut::outDacValidRamSignal(int ch) const {
        return outDacChEnabled(ch) && outDacSignal(ch) && (outDacChMode(ch)==OutChModeRam);
    }

    bool DevOut::outDigValidRamSignal(int ch) const {
        return outDigChEnabled(ch) && outDigSignal(ch) && (outDigChMode(ch)==OutChModeRam);
    }

    bool DevOut::outDacChEnabled(int ch) const {
        return m_dacChEnabled[ch];
    }


    bool DevOut::outDigChEnabled(int ch) const {
        return m_doutChEnabled[ch];
    }


    void DevOut::outDacChEnable(int ch, bool en) {
        m_dacChEnabled[ch] = en;
    }

    void DevOut::outDigChEnable(int ch, bool en) {
        m_doutChEnabled[ch] = en;
    }

    int DevOut::outDacEnabledChCnt() const {
        int ch_cnt = 0;
        for (int ch = 0; ch < outDacChannelsCnt(); ch++) {
            if (outDacValidRamSignal(ch))
                ch_cnt++;
        }
        return ch_cnt;
    }

    int DevOut::outDigEnabledChCnt() const {
        int ch_cnt = 0;
        for (int ch = 0; ch < outDigChannelsCnt(); ch++) {
            if (outDigValidRamSignal(ch))
                ch_cnt++;
        }
        return ch_cnt;
    }

    Error DevOut::outLoadSignalConfig(QSettings &set) {
        if (outDacChannelsCnt() != 0) {
            int size = set.beginReadArray("DacChSignals");
            for (int ch = 0; (ch < outDacChannelsCnt()) && (ch < size); ch++) {
                set.setArrayIndex(ch);
                outDacChEnable(ch, set.value("en").toBool());
                outDacSetSignal(ch, OutSignalFabric::instance()->createDacSignal(set));
            }
            set.endArray();
        }

        if (outDigChannelsCnt() != 0) {
            int size = set.beginReadArray("DoutChSignals");
            for (int ch = 0; (ch < outDigChannelsCnt()) && (ch < size); ch++) {
                set.setArrayIndex(ch);
                outDigChEnable(ch, set.value("en").toBool());
                outDigSetSignal(ch, OutSignalFabric::instance()->createDigSignal(set));
            }
            set.endArray();
        }

        return Error::Success();
    }

    Error DevOut::outSaveSignalConfig(QSettings &set) const {
        if (outDacChannelsCnt() != 0) {
            set.beginWriteArray("DacChSignals", outDacChannelsCnt());
            for (int ch = 0; ch < outDacChannelsCnt(); ch++) {
                set.setArrayIndex(ch);
                QSharedPointer<DacSignal> sig = outDacSignal(ch);
                set.setValue("en", outDacChEnabled(ch));
                if (sig) {
                    OutSignalFabric::instance()->saveSignal(sig, set);
                }
            }
            set.endArray();
        }

        if (outDigChannelsCnt() != 0) {
            set.beginWriteArray("DoutChSignals", outDigChannelsCnt());
            for (int ch = 0; ch < outDigChannelsCnt(); ch++) {
                set.setArrayIndex(ch);
                QSharedPointer<DigSignal> sig = outDigSignal(ch);
                set.setValue("en", outDigChEnabled(ch));
                if (sig) {
                    OutSignalFabric::instance()->saveSignal(sig, set);
                }
            }
            set.endArray();
        }

        return Error::Success();
    }


    int DevOut::outDigChannelsCnt() const {
        return m_sigsDig.size();
    }

    Error DevOut::outSignalsGenSetup() {
        Error err = Error::Success();
        unsigned sig_size;
        err = outCalcCycleSigsParams(outFreq(), outGenSignalMaxSize(), &sig_size);

        if (err.isSuccess())
            err = protOutCycleLoadStart(sig_size);

        if (err.isSuccess()) {
            QVector<double> dac_data;
            QVector<unsigned> dig_data;

            outGenRamInit();
            err = outGenRamData(sig_size, dac_data, dig_data);

            if (err.isSuccess() && ((dac_data.size() + dig_data.size())!=0)) {
                err = protOutSendData(dac_data.data(), dac_data.size(),
                                      dig_data.data(), dig_data.size(),
                                      0, outSendTout(dac_data.size(), dig_data.size()));
            }
        }

        if (err.isSuccess())
            err = protOutFlushData(0);

        if (err.isSuccess())
            err = protOutCycleLoadFinish();        



        if (err.isSuccess()) {
            LQMeasLog->detail(Device::tr("Output cycle generation setup done successfully"), m_dev);
        } else {
            LQMeasLog->error(Device::tr("Output cycle generation setup error"), err, m_dev);
        }
        return err;
    }

    Error DevOut::outSignalsGenStart() {
        Error err = protOutCycleGenStart();
        if (err.isSuccess()) {
            LQMeasLog->detail(Device::tr("Output cycle generation started successfully"), m_dev);
        } else {
            LQMeasLog->error(Device::tr("Output cycle generation start error"), err, m_dev);
        }
        return err;
    }

    Error DevOut::outSignalsGenStop() {
        Error err = protOutCycleGenStop();
        if (err.isSuccess()) {
            LQMeasLog->detail(Device::tr("Output cycle generation stopped successfully"), m_dev);
        } else {
            LQMeasLog->error(Device::tr("Output cycle generation stop error"), err, m_dev);
        }
        return err;
    }

    Error DevOut::outSignalsGenerate() {
        Error err = outSignalsGenSetup();
        if (err.isSuccess())
            err = outSignalsGenStart();
        return err;
    }

    unsigned DevOut::outSendTout(unsigned dac_size, unsigned dout_size)     {
        return 3000 + (dac_size + dout_size)/1000;
    }


    void DevOut::outClearSignals() {
        for (int ch = 0; ch < m_sigsAnalog.size(); ch++) {
            m_sigsAnalog[ch] = QSharedPointer<DacSignal>();
        }
        for (int ch = 0; ch < m_sigsDig.size(); ch++) {
            m_sigsDig[ch] = QSharedPointer<DigSignal>();
        }
    }

    void DevOut::outSetDacChCnt(int ch_cnt) {
        m_sigsAnalog.resize(ch_cnt);
        m_dacChEnabled.resize(ch_cnt);
        for (int ch=0; ch < ch_cnt; ch++)
            m_dacChEnabled[ch] = false;
    }

    void DevOut::outSetDigChCnt(int ch_cnt) {
        m_sigsDig.resize(ch_cnt);
        m_doutChEnabled.resize(ch_cnt);
        for (int ch=0; ch < ch_cnt; ch++)
            m_doutChEnabled[ch] = false;
    }

    Error DevOut::outCalcCycleSigsParams(double out_freq, unsigned max_sig_size,
                                               unsigned *res_size) {
        Error err = Error::Success();
        unsigned sig_size = 1;
        QList<QSharedPointer<OutSignal> > ram_sigs;


        for (int ch = 0; ch < outDacChannelsCnt(); ch++) {
            if (outDacValidRamSignal(ch))
                ram_sigs.append(outDacSignal(ch));
        }

        for (int ch = 0; ch < outDigChannelsCnt(); ch++) {
            if (outDigValidRamSignal(ch))
                ram_sigs.append(outDigSignal(ch));
        }


        foreach (QSharedPointer<OutSignal> sig, ram_sigs) {

            double sig_freq = sig->signalFreq();
            long n, d;

            if (sig_freq <= 0) {
                /* если сигнал имеет явно заданное кол-во точек, а не
                   определяемое по частоте, то нужно установить именно это
                   кол-во отсчетов - как НОК. Если не хватает места - то сигнал
                   не устанавливаем */
                unsigned cur_size = lcm(sig_size, sig->minPoints());
                if (max_sig_size && (cur_size > max_sig_size)) {
                    err = StdErrors::OutSignalPointCount();
                    break;
                }
            } else {
                /* если сигнал задан частотой, то пытаемся
                   подобрать ближайшую к отношению текущей частоты и частоты
                   сигнала дробь */


                /* если задана силшком большая частота - обрезаем до макс. возможной */
                if (out_freq < sig->minPoints()*sig_freq)
                    sig_freq = out_freq/sig->minPoints();


                rat_approx(sig_freq/(out_freq/sig_size), &n, &d, max_sig_size/sig_size,
                           m_cycle_freq_rel_err);
                qDebug() << QString("sig_f = %0, sig_size = %1, n = %2, d = %3")
                            .arg(QString::number(sig_freq))
                            .arg(QString::number(sig_size))
                            .arg(QString::number(n))
                            .arg(QString::number(d));
                /* нулевой числитель => доступного буфера не достаточно даже для одного
                   периода */
                if (n==0) {
                    err = StdErrors::OutSignalPointCount();
                    break;
                }
                sig_size*=d;
            }
        }

        if (err.isSuccess()) {
            foreach (QSharedPointer<OutSignal> sig, ram_sigs) {
                if (sig->signalFreq() > 0) {
                    if (!sig->periodStrictPoints()) {
                        unsigned periods = sig->signalFreq()*sig_size/out_freq + 0.5;
                        if (periods==0)
                            periods = 1;
                        sig->setRealParams(out_freq*periods/sig_size, out_freq);
                    } else {
                        unsigned req_size = (unsigned)(out_freq/sig->signalFreq()+0.5);
                        unsigned periods = (double)sig_size/req_size + 0.5;
                        if (periods==0)
                            periods = 1;
                        req_size = sig_size/periods;
                        sig->setRealParams(out_freq/req_size, out_freq);
                    }
                } else {
                    sig->setRealParams(OutSignal::unspecifiedFreq(), out_freq);
                }
            }
        }

        if (err.isSuccess())
            *res_size = sig_size;

        return err;
    }
}
