#include "MultimeterThread.h"
#include <QMetaType>
#include <QElapsedTimer>
#include "../../devs/LQMeasDevice.h"
#include "../../lqmeas_errs.h"
#include "lmath.h"
#include "MultimeterProcessor.h"

MultimeterThread::MultimeterThread(QObject *parent) :
    QThread(parent), m_cfgReq(false), m_stopReq(false), m_cur_id(0),
    m_curCfg(0), m_cfgSize(0) {

    m_proc = new MultimeterProcessor();




    m_proc->moveToThread(&m_procThread);

    qRegisterMetaType<MultimeterData>("MultimeterThread::MultimeterData");
    qRegisterMetaType<QVector<double> >("QVector<double>");
    connect(m_proc, SIGNAL(dataProcessed(MultimeterThread::MultimeterData)),
            SIGNAL(dataReceived(MultimeterThread::MultimeterData)),  Qt::DirectConnection);
    connect(this, SIGNAL(dataForProcess(LQMeasAdc*,QVector<double>,MultimeterThread::MultimeterData)),
            m_proc, SLOT(processData(LQMeasAdc*,QVector<double>,MultimeterThread::MultimeterData)));
}

void MultimeterThread::setDev(QSharedPointer<LQMeasAdc> adc) {
    m_dev = qSharedPointerDynamicCast<LQMeasDevice>(adc);
    m_adc = adc.data();
}

void MultimeterThread::run() {
    bool out = false;
    LQMeasError err = LQMeasError::Success();
    int meas_time;
    int adc_ch_cnt = m_adc->adcChannelsCnt();
    int recv_size, ch_size;


    MultimeterData data;
    data.ch.resize(adc_ch_cnt);

    m_procThread.start();


    m_cfgReq = true;
    m_stopReq = false;
    m_err = LQMeasError::Success();

    m_dev->reportMessage(QxtLogger::InfoLevel, tr("Запуск измерений"));


    while (!out && err.isSuccess()) {
        m_lock.lock();
        /* проверка изменений настроек */
        if (m_cfgReq) {           

            /* останавливаем сбор, если он был запущен */
            if (m_adc->adcIsRunning())
                err = m_adc->adcStop();

            if (err.isSuccess()) {
                m_dev->reportMessage(QxtLogger::InfoLevel, tr("Обновление настроек модуля"));

                err = m_adc->adcSetRawConfig(m_curCfg, m_cfgSize);
                if (err.isSuccess()) {
                    err = m_dev->configure();
                }

                m_cfgReq = false;
                meas_time = m_meas_time;
                data.cfgID = m_cur_id;

                if (err.isSuccess()) {
                    ch_size = meas_time*m_adc->adcFreq()/1000.;
#ifndef LQMEAS_SERVICE_MULTIMETER_DC_ONLY
                    if (ch_size < 2)
                        ch_size = 2;
#endif

                    recv_size = ch_size*m_adc->adcEnabledChCnt();
                }
            }

            if (err.isSuccess())
                err = m_adc->adcStart();
        }

        if (m_stopReq) {
            out = true;
            m_stopReq = false;
        }

        m_lock.unlock();

        QVector<double> vals;
        vals.resize(recv_size);


        while (!out && err.isSuccess() && !restartReq()) {
            int tout = meas_time+4000;
            QElapsedTimer tmr;
            int recvd_size = 0;
            tmr.start();

            while (err.isSuccess() && !restartReq() && (recvd_size!=recv_size) && !tmr.hasExpired(tout)) {
                unsigned cur_recvd_size=0;
                err = m_adc->adcGetData(&vals.data()[recvd_size], recv_size-recvd_size, 0,
                                     500, &cur_recvd_size);
                if (err.isSuccess())
                    recvd_size+=cur_recvd_size;
            }

            if (err.isSuccess() && !restartReq() && (recvd_size!=recv_size)) {
                err = LQMeasStdErrors::RecvInsufficientData();
                qxtLog->error("Не удалось дождаться всего кадра за заданное время");
            }

            if (err.isSuccess() && !restartReq()) {


                /* раскладываем данные по каналам */
                for (int ch=0; ch < adc_ch_cnt; ch++) {
                    data.ch[ch].status = m_adc->adcGetChStatus(ch, true);
                }
                emit dataForProcess(m_adc, vals, data);

            }
        }
    }

    if (m_adc->adcIsRunning()) {
        LQMeasError stop_err = m_adc->adcStop();
        if (err.isSuccess())
            err = stop_err;
    }

    m_err = err;

    m_procThread.quit();


    m_dev->reportMessage(QxtLogger::InfoLevel, tr("Измерения остановлены"));
}

LQMeasError MultimeterThread::lastError() const {
    return m_err;
}

void MultimeterThread::setConfig(void *cfg, size_t cfg_size, int meas_time, unsigned *id) {
    QMutexLocker locker(&m_lock);
    if (m_cfgSize != cfg_size) {
        delete [] m_curCfg;
        m_curCfg = new unsigned char[cfg_size];
        m_cfgSize = cfg_size;
    }
    memcpy(m_curCfg, cfg, m_cfgSize);
    m_cfgReq = true;
    m_meas_time = meas_time;
    m_cur_id++;
    if (id!=0)
        *id = m_cur_id;
}

void MultimeterThread::stopRequest() {
    QMutexLocker locker(&m_lock);
    m_stopReq = true;
}



