#include "X502ConfigDialog.h"
#include "ui_X502ConfigDialog.h"

#include <QTableWidget>
#include <QTableWidgetItem>
#include <QHeaderView>
#include <QComboBox>
#include <QSpinBox>
#include <QMessageBox>
#include <QDialogButtonBox>
#include <QPushButton>
#include "QLedIndicator.h"

#include "out/OutSignalConfigCtl.h"

#include "lqmeas/devs/x502/e502.h"
#include "lqmeas/devs/x502/l502.h"

X502ConfigDialog::X502ConfigDialog(QSharedPointer<LQMeas::X502> dev, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::X502ConfigDialog),
    m_dev(dev) {

    bool change_enabled = !m_dev->adcIsRunning();

    ui->setupUi(this);

    setWindowTitle(windowTitle() + ". " + dev->devStrWithLocation());

    ui->serial->setText(dev->serial());
    ui->pldaVer->setText(QString::number(dev->pldaVer()));
    ui->fpgaVer->setText(QString("%0.%1").arg(QString::number((dev->fpgaVer() >> 8) & 0xFF))
                         .arg(QString::number(dev->fpgaVer()&0xFF)));
    if (dev->name() == LQMeas::L502::typeDeviceName()) {
        ui->armVer->setVisible(false);
        ui->armVerLbl->setVisible(false);
    } else {
        ui->armVer->setText(QString("%0.%1.%2.%3")
                            .arg(QString::number((dev->mcuVer() >> 24) & 0xFF))
                            .arg(QString::number((dev->mcuVer() >> 16) & 0xFF))
                            .arg(QString::number((dev->mcuVer() >>  8) & 0xFF))
                            .arg(QString::number(dev->mcuVer() & 0xFF)));
    }

    addOption(tr("DAC"), dev->dacPresent());
    addOption(tr("Galvanic decoupling"), dev->galPresent());
    addOption(tr("BlackFin"), dev->bfPresent());
    if (dev->name() == LQMeas::E502::typeDeviceName()) {
        addOption(tr("Ethernet"), dev->ethPresent());
    }

    LQMeas::X502::AdcConfig cfg = dev->adcCurConfig();
    unsigned ch_cnt = dev->adcChannelsCnt();

    ui->adcChannels->setRowCount(ch_cnt);


    for (unsigned ch=0; ch < ch_cnt; ch++) {
        ChUi ch_ui;
        ch_ui.chItem = new QTableWidgetItem();
        ch_ui.chItem->setText(tr("Channel %1").arg(QString::number(ch+1)));
        ch_ui.chItem->setCheckState(cfg.Ch[ch].Enabled ? Qt::Checked : Qt::Unchecked);
        if (!change_enabled)
            ch_ui.chItem->setFlags(ch_ui.chItem->flags() &= ~Qt::ItemIsEnabled);


        ch_ui.mode = new QComboBox();
        ch_ui.mode->addItem(tr("Comm. mode"), X502_LCH_MODE_COMM);
        if (ch < X502_ADC_DIFF_CH_CNT) {
            ch_ui.mode->addItem(tr("Diff."), X502_LCH_MODE_DIFF);
        }
        ch_ui.mode->addItem(tr("Zero"), X502_LCH_MODE_ZERO);
        ch_ui.mode->setEnabled(change_enabled);
#ifndef _WIN32
        ch_ui.mode->setStyleSheet("QComboBox { padding: 1px 18px 1px 0px; }");
#endif

        ch_ui.range = new QComboBox();
        ch_ui.range->addItem(tr("± 10 V"),  X502_ADC_RANGE_10);
        ch_ui.range->addItem(tr("± 5 V"),   X502_ADC_RANGE_5);
        ch_ui.range->addItem(tr("± 2 V"),   X502_ADC_RANGE_2);
        ch_ui.range->addItem(tr("± 1 V"),   X502_ADC_RANGE_1);
        ch_ui.range->addItem(tr("± 0.5 V"), X502_ADC_RANGE_05);
        ch_ui.range->addItem(tr("± 0.2 V"), X502_ADC_RANGE_02);
        ch_ui.range->setEnabled(change_enabled);
#ifndef _WIN32
        ch_ui.range->setStyleSheet("QComboBox { padding: 1px 18px 1px 0px; }");
#endif

        ch_ui.avg = new QSpinBox();
        ch_ui.avg->setMinimum(1);
        ch_ui.avg->setMaximum(255);
        ch_ui.avg->setSpecialValueText(tr("Off"));
        ch_ui.avg->setReadOnly(!change_enabled);
#ifndef _WIN32
        ch_ui.avg->setStyleSheet("* { padding: 1px 18px 1px 0px; }");
#endif






        setComboBoxItemByData(ch_ui.mode,  cfg.Ch[ch].Mode);
        setComboBoxItemByData(ch_ui.range, cfg.Ch[ch].Range);
        ch_ui.avg->setValue(cfg.Ch[ch].Avg);


        ui->adcChannels->setItem(ch, 0, ch_ui.chItem);
        ui->adcChannels->setCellWidget(ch, 1, ch_ui.mode);
        ui->adcChannels->setCellWidget(ch, 2, ch_ui.range);
        ui->adcChannels->setCellWidget(ch, 3, ch_ui.avg);

        m_chUis.append(ch_ui);
    }



    ui->adcChannels->resizeColumnsToContents();

    ui->adcFreq->setValue(dev->adcFreq());
    ui->adcChFreq->setValue(dev->adcChFreq());
    if (cfg.FreqCh == 0) {
        ui->adcMaxChFreqBox->setCheckState(Qt::Checked);
        ui->adcChFreq->setEnabled(false);
    } else {
        ui->adcMaxChFreqBox->setCheckState(Qt::Unchecked);
        ui->adcChFreq->setEnabled(true);
    }
    connect(ui->adcMaxChFreqBox, SIGNAL(stateChanged(int)), SLOT(adcMaxInChFreqChanged(int)));


    ui->adcFreq->setReadOnly(!change_enabled);
    ui->adcChFreq->setReadOnly(!change_enabled);
    ui->adcMaxChFreqBox->setEnabled(change_enabled);


    QSettings set;
    set.beginGroup("X502CfgDialog");
    ui->adcFreq->loadUnitSettings(set, "AdcFreqUnit");
    ui->outFreq->loadUnitSettings(set, "OutFreqUnit");

    set.endGroup();

    ui->adcChFreq->setUnitMultCoef(ui->adcFreq->unitMultCoef());

    if (change_enabled) {
        ui->adcChFreq->setUnitReadOnly(true);
        connect(ui->adcFreq, SIGNAL(unitChanged(int)), ui->adcChFreq, SLOT(setUnitMultCoef(int)));
    }


    fillSyncBox(ui->syncFreqMode, dev->syncFreqMode());
    fillSyncBox(ui->syncStartMode, dev->syncStartMode());

    ui->syncFreqMode->setEnabled(change_enabled);
    ui->syncStartMode->setEnabled(change_enabled);


    ui->refFreq->addItem("2000000" + tr(" Hz"), X502_REF_FREQ_2000KHZ);
    ui->refFreq->addItem("1500000" + tr(" Hz"), X502_REF_FREQ_1500KHZ);
    t_x502_ref_freq freq = dev->refFreq();
    setComboBoxItemByData(ui->refFreq, (int)freq);
    ui->refFreq->setEnabled(change_enabled);

    uint32_t pullups = dev->pullups();


    if (dev->name() != LQMeas::L502::typeDeviceName()) {
        ui->pullLow->setVisible(false);
        ui->pullHigh->setVisible(false);
    } else {
        ui->pullLow->setChecked(pullups & X502_PULLUPS_DI_L);
        ui->pullHigh->setChecked(pullups & X502_PULLUPS_DI_H);

        ui->pullLow->setEnabled(change_enabled);
        ui->pullHigh->setEnabled(change_enabled);
    }

    if (dev->name() != LQMeas::E502::typeDeviceName()) {
        ui->pullCONV->setVisible(false);
        ui->pullSTART->setVisible(false);
    } else {
        ui->pullCONV->setChecked(pullups & X502_PULLDOWN_CONV_IN);
        ui->pullSTART->setChecked(pullups & X502_PULLDOWN_START_IN);

        ui->pullCONV->setEnabled(change_enabled);
        ui->pullSTART->setEnabled(change_enabled);
    }
    ui->pullSYN1->setChecked(pullups & X502_PULLUPS_DI_SYN1);
    ui->pullSYN2->setChecked(pullups & X502_PULLUPS_DI_SYN2);

    ui->pullSYN1->setEnabled(change_enabled);
    ui->pullSYN2->setEnabled(change_enabled);

    ui->outFreq->setValue(dev->outFreq());

    if ((dev->name() == LQMeas::E502::typeDeviceName())
            && dev->dacPresent()) {
        ui->outFreq->setReadOnly(!change_enabled);        
    } else {
        ui->outFreq->setEnabled(false);
    }
    ui->outMode->addItem(tr("Cycle"), LQMeas::DevOut::OutRamSyncModeCycle);
    ui->outMode->addItem(tr("Stream"), LQMeas::DevOut::OutRamSyncModeStream);

    setComboBoxItemByData(ui->outMode, dev->outRamSyncMode());

    ui->outMode->setEnabled(dev->dacPresent() && change_enabled);





    new OutSignalConfigCtl(this, dev.data(),
                           QList<QCheckBox*>() << ui->dacCh1En << ui->dacCh2En,
                           QList<QLineEdit*>() << ui->dacCh1SigEdit << ui->dacCh2SigEdit,
                           change_enabled);


    ui->buttonBox->button(QDialogButtonBox::Apply)->setEnabled(change_enabled);
    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(change_enabled);
    connect(ui->buttonBox->button(QDialogButtonBox::Apply), SIGNAL(clicked()),
            SLOT(setConfig()));
    connect(ui->buttonBox->button(QDialogButtonBox::Cancel), SIGNAL(clicked()),
            SLOT(reject()));

    /* нельзя использовать каналы 17-32, если разрешены соответствующие
       каналы 1-16 в диф. режиме. Отслеживаем это динамически */
    if (change_enabled) {
        for (int ch=0; ch < X502_ADC_DIFF_CH_CNT; ch++) {
            updateCommChState(ch);
            connect(m_chUis[ch].mode, SIGNAL(currentIndexChanged(int)), SLOT(chModeChanged()));
        }
        connect(ui->adcChannels, SIGNAL(cellChanged(int,int)), SLOT(cellDataChanged(int,int)));

        /* при изменении коэф. усреднения изменяем соответствующим образом макс.
           частоту сбора */
        for (int ch=0; ch < m_chUis.size(); ch++) {
            connect(m_chUis[ch].avg, SIGNAL(valueChanged(int)), SLOT(setMaxAdcFreq()));
        }
        connect(ui->adcChannels, SIGNAL(cellChanged(int,int)), SLOT(setMaxAdcFreq()));
        connect(ui->refFreq, SIGNAL(currentIndexChanged(int)), SLOT(setMaxAdcFreq()));
        connect(ui->adcFreq, SIGNAL(editingFinished()), SLOT(setMaxAdcChFreq()));
        setMaxAdcFreq();
    }
}

X502ConfigDialog::~X502ConfigDialog() {
    delete ui;
}

void X502ConfigDialog::accept() {
    if (setConfig().isSuccess())
        QDialog::accept();
}

void X502ConfigDialog::setComboBoxItemByData(QComboBox *box, int data) {
    int idx = 0;
    for (int i=0; i < box->count(); i++) {
        if (box->itemData(i).toInt()==data) {
            idx = i;
            break;
        }
    }
    box->setCurrentIndex(idx);
}

void X502ConfigDialog::fillSyncBox(QComboBox *box, int value) {
    box->addItem(tr("Internal"), X502_SYNC_INTERNAL);
    box->addItem(tr("From external master"), X502_SYNC_EXTERNAL_MASTER);
    box->addItem(tr("SYN1 Rise"), X502_SYNC_DI_SYN1_RISE);
    box->addItem(tr("SYN2 Rise"), X502_SYNC_DI_SYN2_RISE);
    box->addItem(tr("SYN1 Fall"), X502_SYNC_DI_SYN1_FALL);
    box->addItem(tr("SYN2 Fall"), X502_SYNC_DI_SYN2_FALL);
    setComboBoxItemByData(box, value);
}

LQMeas::Error X502ConfigDialog::setConfig() {
    LQMeas::Error err;
    LQMeas::X502::AdcConfig cfg = m_dev->adcCurConfig();
    unsigned ch_cnt = m_dev->adcChannelsCnt();

    /* сохраняем выбранные единицы измерения */
    QSettings set;
    set.beginGroup("X502CfgDialog");
    ui->adcFreq->saveUnitSettings(set, "AdcFreqUnit");
    ui->outFreq->saveUnitSettings(set, "OutFreqUnit");
    set.endGroup();


    for (unsigned ch=0; ch < ch_cnt; ch++) {
        cfg.Ch[ch].Enabled = m_chUis[ch].chItem->checkState() == Qt::Checked;
        cfg.Ch[ch].Mode    = (t_x502_lch_mode)m_chUis[ch].mode->itemData(m_chUis[ch].mode->currentIndex()).toInt();
        cfg.Ch[ch].Range   = (t_x502_adc_range)m_chUis[ch].range->itemData(m_chUis[ch].range->currentIndex()).toInt();
        cfg.Ch[ch].Avg     = m_chUis[ch].avg->value();
    }

    cfg.FreqAdc = ui->adcFreq->value();
    cfg.FreqCh = ui->adcMaxChFreqBox->isChecked() ? 0 : ui->adcChFreq->value();

    err = m_dev->adcSetConfig(&cfg);
    if (err.isSuccess())
        err = m_dev->setOutFreq(ui->outFreq->value());
    if (err.isSuccess())
        m_dev->outSetRamSyncMode((LQMeas::DevOut::OutRamSyncMode)ui->outMode->itemData(ui->outMode->currentIndex()).toInt());
    if (err.isSuccess())
        err = m_dev->setSyncFreqMode((t_x502_sync_mode)ui->syncFreqMode->itemData(ui->syncFreqMode->currentIndex()).toInt());
    if (err.isSuccess())
        err = m_dev->setSyncStartMode((t_x502_sync_mode)ui->syncStartMode->itemData(ui->syncStartMode->currentIndex()).toInt());
    if (err.isSuccess())
        err = m_dev->setRefFreq((t_x502_ref_freq)ui->refFreq->itemData(ui->refFreq->currentIndex()).toInt());
    if (err.isSuccess()) {
        uint32_t pullups = 0;
        if (ui->pullLow->isChecked())
            pullups |= X502_PULLUPS_DI_L;
        if (ui->pullHigh->isChecked())
            pullups |= X502_PULLUPS_DI_H;
        if (ui->pullCONV->isChecked())
            pullups |= X502_PULLDOWN_CONV_IN;
        if (ui->pullSTART->isChecked())
            pullups |= X502_PULLDOWN_START_IN;
        if (ui->pullSYN1->isChecked())
            pullups |= X502_PULLUPS_DI_SYN1;
        if (ui->pullSYN2->isChecked())
            pullups |= X502_PULLUPS_DI_SYN2;

        err = m_dev->setPullups(pullups);
    }



    if (err.isSuccess())
        err = m_dev->configure();


    if (err.isSuccess()) {
        ui->adcFreq->setValue(m_dev->adcFreq());
        ui->adcChFreq->setValue(m_dev->adcChFreq());
        ui->outFreq->setValue(m_dev->outFreq());
    } else {
        QMessageBox::critical(this, tr("ADC configuration error"), err.msg());
    }
    return err;
}

void X502ConfigDialog::adcMaxInChFreqChanged(int val) {
    ui->adcChFreq->setEnabled(val != Qt::Checked);
}

void X502ConfigDialog::chModeChanged() {
    QComboBox *modebox = qobject_cast<QComboBox *>(sender());
    for (int i=0; i < X502_ADC_DIFF_CH_CNT; i++) {
        if (m_chUis[i].mode == modebox) {
            updateCommChState(i);
            break;
        }
    }
}

void X502ConfigDialog::updateCommChState(int i) {
    bool com_ch_en = (m_chUis[i].chItem->checkState() != Qt::Checked) ||
            (m_chUis[i].mode->itemData(m_chUis[i].mode->currentIndex()).toInt() != X502_LCH_MODE_DIFF);
    QTableWidgetItem *com_ch_item = m_chUis[i + X502_ADC_DIFF_CH_CNT].chItem;
    if (com_ch_en) {
        com_ch_item->setFlags(com_ch_item->flags() | Qt::ItemIsEnabled);
    } else {
        com_ch_item->setFlags(com_ch_item->flags() &= ~Qt::ItemIsEnabled);
        com_ch_item->setCheckState(Qt::Unchecked);
    }
    m_chUis[i + X502_ADC_DIFF_CH_CNT].avg->setEnabled(com_ch_en);
    m_chUis[i + X502_ADC_DIFF_CH_CNT].mode->setEnabled(com_ch_en);
    m_chUis[i + X502_ADC_DIFF_CH_CNT].range->setEnabled(com_ch_en);
}

void X502ConfigDialog::cellDataChanged(int row, int col) {
    if ((col == 0) && (row < X502_ADC_DIFF_CH_CNT))
        updateCommChState(row);
}

void X502ConfigDialog::setMaxAdcFreq() {
    int max_avg = 1;
    for (int ch = 0; ch < m_chUis.size(); ch++) {
        if (m_chUis[ch].chItem->checkState() == Qt::Checked) {
            int avg = m_chUis[ch].avg->value();
            if (avg > max_avg)
                max_avg = avg;
        }
    }
    ui->adcFreq->setMaximum(ui->refFreq->itemData(ui->refFreq->currentIndex()).toDouble()/max_avg);
    setMaxAdcChFreq();
}

void X502ConfigDialog::setMaxAdcChFreq() {
    int ch_cnt = 0;
    for (int ch = 0; ch < m_chUis.size(); ch++) {
        if (m_chUis[ch].chItem->checkState() == Qt::Checked) {
            ch_cnt++;
        }
    }
    if (ch_cnt != 0) {
        ui->adcChFreq->setMaximum(ui->adcFreq->value()/ch_cnt);
        if (ui->adcMaxChFreqBox->isChecked())
            ui->adcChFreq->setValue(ui->adcFreq->value()/ch_cnt);
    }

}

void X502ConfigDialog::addOption(QString text, bool val) {
    QLedIndicator *led = new QLedIndicator(0, 22);
    led->setEnabled(false);
    led->setChecked(val);
    ui->optionsLout->addRow(text, led);
}

